/******************************************************************************
 * @brief    wifi(AT-commandʾ, ʹõģM169WI-FI)
 *
 * Copyright (c) 2020, <morro_luo@163.com>
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Change Logs: 
 * Date           Author       Notes 
 * 2021-01-20     Morro        
 * 2021-03-03     Morro        URCʹð
 ******************************************************************************/
#include "at_chat.h"
#include "wifi_uart.h"
#include "public.h"
#include "module.h"
#include "cli.h"
#include <stdio.h>
#include <stdbool.h>

/* Private function prototypes -----------------------------------------------*/
void wifi_open(void);
void wifi_close(void);
static void at_error(void);
void wifi_query_version(void);
void wifi_ready_handler(char *recvbuf, int size);
void wifi_connected_handler(char *recvbuf, int size);
void wifi_disconnected_handler(char *recvbuf, int size);

/* Private variables ---------------------------------------------------------*/
/* 
 * @brief   AT
 */
static at_obj_t at;

/* 
 * @brief   wifi ݽջ
 */
static unsigned char wifi_recvbuf[256];

/* 
 * @brief   wifi URCջ
 */
static unsigned char wifi_urcbuf[128];

/* 
 * @brief   wifi URC
 */
static const utc_item_t urc_table[] = {
    "ready",             wifi_ready_handler,
    "WIFI CONNECTED:",   wifi_connected_handler,
    "WIFI DISCONNECTED", wifi_disconnected_handler,
};

/* 
 * @brief   AT
 */
static const at_adapter_t  at_adapter = {
    .write    = wifi_uart_write,
    .read     = wifi_uart_read,
    .error    = at_error,
    .utc_tbl  = (utc_item_t *)urc_table,
    .urc_buf  = wifi_urcbuf,
    .recv_buf = wifi_recvbuf,
    .urc_tbl_count = sizeof(urc_table) / sizeof(urc_table[0]),
    .urc_bufsize   = sizeof(wifi_urcbuf),
    .recv_bufsize = sizeof(wifi_recvbuf)
};

/* Private functions ---------------------------------------------------------*/

/* 
 * @brief   wifi¼
 */
void wifi_ready_handler(char *recvbuf, int size)
{
    printf("WIFI ready...\r\n");
}

/* 
 * @brief   wifi¼
 */
static void wifi_connected_handler(char *recvbuf, int size)
{
    printf("WIFI connection detected...\r\n");
}
/* 
 * @brief   wifiϿ¼
 */
static void wifi_disconnected_handler(char *recvbuf, int size)
{
    printf("WIFI disconnect detected...\r\n");
}

/* 
 * @brief   wifi
 */
void wifi_open(void)
{
    GPIO_SetBits(GPIOA, GPIO_Pin_4);
    printf("wifi open\r\n");
}
/* 
 * @brief   رwifi
 */
void wifi_close(void)
{
    GPIO_ResetBits(GPIOA, GPIO_Pin_4);
    printf("wifi close\r\n");
}


/* 
 * @brief   WIFI״̬
 * @return  true - ˳״̬, false - ״̬,
 */
static int wifi_reset_work(at_env_t *e)
{
    at_obj_t *a = (at_obj_t *)e->params;
    switch (e->state) {
    case 0:                                //رWIFIԴ
        wifi_close();
        e->reset_timer(a);
        e->state++;
        break;
    case 1:
        if (e->is_timeout(a, 2000))       //ʱȴ2s
            e->state++;
        break;
    case 2:
        wifi_open();                       //wifi
        e->state++;
        break;
    case 3:
        if (e->is_timeout(a, 5000))       //Լʱȴ5swifi
            return true;  
        break;
    }
    return false;
}
/* 
 * @brief   wifi ͨ쳣
 */
static void at_error(void)
{
    printf("wifi AT communication error\r\n");
    //ִҵ
    at_do_work(&at, wifi_reset_work, &at);        
}

/* 
 * @brief    ʼص
 */
static void at_init_callbatk(at_response_t *r)
{    
    if (r->ret == AT_RET_OK ) {
        printf("wifi Initialization successfully...\r\n");
        
        /* ѯ汾*/
        wifi_query_version();
        
    } else 
        printf("wifi Initialization failure...\r\n");
}

/* 
 * @brief    wifiʼ
 */
static const char *wifi_init_cmds[] = {
    "AT+GPIO_WR=1,1\r\n",
    "AT+GPIO_WR=2,0\r\n",
    "AT+GPIO_WR=3,1\r\n",
    NULL
};


/* 
 * @brief    wifiʼ
 */
void wifi_init(void)
{
    RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOA , ENABLE);
    gpio_conf(GPIOA, GPIO_Mode_OUT, GPIO_PuPd_NOPULL, GPIO_Pin_4);
    
    wifi_uart_init(115200);
    at_obj_init(&at, &at_adapter);
         
    //WIFI
    at_do_work(&at, wifi_reset_work, &at);        
    
    //ʼwifi
    at_send_multiline(&at, at_init_callbatk, wifi_init_cmds);  
    
    //GPIO
    at_send_singlline(&at, NULL, "AT+GPIO_TEST_EN=1\r\n");  
    
}driver_init("wifi", wifi_init); 



/* 
 * @brief    wifi(10ms ѯ1)
 */
void wifi_task(void)
{
    at_poll_task(&at);
}task_register("wifi", wifi_task, 10);


/** Ǳ׼AT----------------------------------------------------------------
 *  Բѯ汾Ϊ:
 * ->  AT+VER\r\n
 * <-  VERSION:M169-YH01
 *  
 */

//ʽ1, ʹat_do_cmdӿ

/* 
 * @brief    ԶAT
 */
static void at_ver_sender(at_env_t *e)
{
    e->printf(&at, "AT+VER\r\n");
}
/* 
 * @brief    汾ѯص
 */
static void query_version_callback(at_response_t *r)
{
    if (r->ret == AT_RET_OK ) {
        printf("wifi version info : %s\r\n", r->recvbuf);
    } else 
        printf("wifi version query failure...\r\n");
}

/* AT*/
static const at_cmd_t at_cmd_ver = {
    at_ver_sender,                   //ԶAT
    "VERSION:",                      //ƥǰ׺
    query_version_callback,          //ѯص
    3, 3000                          //Դʱʱ
};

/* 
 * @brief    ִа汾ѯ
 */
void wifi_query_version(void)
{
    at_do_cmd(&at, NULL, &at_cmd_ver);
}
